import numpy as np

from biorobot.toy_example.mjcf.morphology.specification.specification import (
    ToyExampleActuationSpecification,
    ToyExampleArmSegmentSpecification,
    ToyExampleArmSpecification,
    ToyExampleJointSpecification,
    ToyExampleMorphologySpecification,
    ToyExampleTorsoSpecification,
)


def default_joint_specification() -> ToyExampleJointSpecification:
    joint_specification = ToyExampleJointSpecification(
        range=20 / 180 * np.pi, stiffness=0.1, damping=0.1
    )
    return joint_specification


def default_actuation_specification() -> ToyExampleActuationSpecification:
    actuation_specification = ToyExampleActuationSpecification(kp=10)
    return actuation_specification


def default_arm_segment_specification() -> ToyExampleArmSegmentSpecification:
    joint_specification = default_joint_specification()
    segment_specification = ToyExampleArmSegmentSpecification(
        radius=0.025, length=0.1, joint_specification=joint_specification
    )
    return segment_specification


def default_arm_specification(num_segments: int) -> ToyExampleArmSpecification:
    segment_specifications = []
    for _ in range(num_segments):
        segment_specification = default_arm_segment_specification()
        segment_specifications.append(segment_specification)

    arm_specification = ToyExampleArmSpecification(
        segment_specifications=segment_specifications
    )
    return arm_specification


def default_torso_specification() -> ToyExampleTorsoSpecification:
    torso_specification = ToyExampleTorsoSpecification(radius=0.1)
    return torso_specification


def default_toy_example_morphology_specification(
    num_arms: int, num_segments_per_arm: int
) -> ToyExampleMorphologySpecification:

    torso_specification = default_torso_specification()

    arm_specifications = []
    for _ in range(num_arms):
        arm_specification = default_arm_specification(num_segments=num_segments_per_arm)
        arm_specifications.append(arm_specification)

    actuation_specification = default_actuation_specification()

    morphology_specification = ToyExampleMorphologySpecification(
        torso_specification=torso_specification,
        arm_specifications=arm_specifications,
        actuation_specification=actuation_specification,
    )

    return morphology_specification



from typing import List

from fprs.parameters import FixedParameter
from fprs.specification import MorphologySpecification, Specification


class ToyExampleJointSpecification(Specification):
    def __init__(self, range: float, stiffness: float, damping: float) -> None:
        super().__init__()
        self.stiffness = FixedParameter(value=stiffness)
        self.range = FixedParameter(value=range)
        self.damping = FixedParameter(value=damping)


class ToyExampleArmSegmentSpecification(Specification):
    def __init__(
        self,
        radius: float,
        length: float,
        joint_specification: ToyExampleJointSpecification,
    ) -> None:
        super().__init__()
        self.radius = FixedParameter(radius)
        self.length = FixedParameter(length)
        self.joint_specification = joint_specification


class ToyExampleArmSpecification(Specification):
    def __init__(
        self, segment_specifications: List[ToyExampleArmSegmentSpecification]
    ) -> None:
        super().__init__()
        self.segment_specifications = segment_specifications


class ToyExampleTorsoSpecification(Specification):
    def __init__(self, radius: float) -> None:
        super().__init__()
        self.radius = FixedParameter(radius)


class ToyExampleActuationSpecification(Specification):
    def __init__(self, kp: float) -> None:
        super().__init__()
        self.kp = FixedParameter(kp)


class ToyExampleMorphologySpecification(MorphologySpecification):
    def __init__(
        self,
        torso_specification: ToyExampleTorsoSpecification,
        arm_specifications: List[ToyExampleArmSpecification],
        actuation_specification: ToyExampleActuationSpecification,
    ) -> None:
        super().__init__()
        self.torso_specification = torso_specification
        self.arm_specifications = arm_specifications
        self.actuation_specification = actuation_specification



from __future__ import annotations

from typing import Any, Dict, List

import chex
import jax.random
from jax import numpy as jnp
from jax._src.scipy.spatial.transform import Rotation
from mujoco._structs import _MjModelJointViews
from moojoco.environment.mjx_env import MJXEnv, MJXEnvState, MJXObservable

from biorobot.toy_example.environment.mjc_env import ToyExampleEnvironmentConfiguration
from biorobot.toy_example.mjcf.arena.arena import MJCFPlaneWithTargetArena
from biorobot.toy_example.mjcf.morphology.morphology import MJCFToyExampleMorphology


class ToyExampleMJXEnvironment(MJXEnv):
    metadata = {"render_modes": ["human", "rgb_array"]}

    def __init__(
        self,
        configuration: ToyExampleEnvironmentConfiguration,
        mjcf_str: str,
        mjcf_assets: Dict[str, Any],
    ) -> None:
        super().__init__(
            mjcf_str=mjcf_str, mjcf_assets=mjcf_assets, configuration=configuration
        )
        self._segment_joints: List[_MjModelJointViews] = [
            self.frozen_mj_model.joint(joint_id)
            for joint_id in range(self.frozen_mj_model.njnt)
            if "segment" in self.frozen_mj_model.joint(joint_id).name
        ]
        self._segment_joint_qpos_adrs = jnp.array(
            [joint.qposadr[0] for joint in self._segment_joints]
        )
        self._segment_joint_qvel_adrs = jnp.array(
            [joint.dofadr[0] for joint in self._segment_joints]
        )

    @classmethod
    def from_morphology_and_arena(
        cls,
        morphology: MJCFToyExampleMorphology,
        arena: MJCFPlaneWithTargetArena,
        configuration: ToyExampleEnvironmentConfiguration,
    ) -> ToyExampleMJXEnvironment:
        return super().from_morphology_and_arena(
            morphology=morphology, arena=arena, configuration=configuration
        )

    @property
    def environment_configuration(self) -> ToyExampleEnvironmentConfiguration:
        return super().environment_configuration

    @staticmethod
    def _get_xy_direction_to_target(
        state: MJXEnvState,
    ) -> jnp.ndarray:
        torso_body_id = state.mj_model.body("ToyExampleMorphology/torso").id
        target_body_id = state.mj_model.body("target").id

        torso_body_pos = state.mjx_data.xpos[torso_body_id]
        target_pos = state.mjx_data.xpos[target_body_id]

        return (target_pos - torso_body_pos)[:2]

    @staticmethod
    def _get_xy_distance_to_target(state: MJXEnvState) -> float:
        xy_direction_to_target = ToyExampleMJXEnvironment._get_xy_direction_to_target(
            state=state
        )
        return jnp.linalg.norm(xy_direction_to_target)

    def _create_observables(self) -> List[MJXObservable]:
        in_plane_joints = [
            self.frozen_mj_model.joint(joint_id)
            for joint_id in range(self.frozen_mj_model.njnt)
            if "in_plane_joint" in self.frozen_mj_model.joint(joint_id).name
        ]
        out_of_plane_joints = [
            self.frozen_mj_model.joint(joint_id)
            for joint_id in range(self.frozen_mj_model.njnt)
            if "out_of_plane_joint" in self.frozen_mj_model.joint(joint_id).name
        ]

        in_plane_joint_qpos_adr = jnp.array(
            [joint.qposadr[0] for joint in in_plane_joints]
        )
        in_plane_joint_range = jnp.array([joint.range for joint in in_plane_joints]).T
        out_of_plane_joint_qpos_adr = jnp.array(
            [joint.qposadr[0] for joint in out_of_plane_joints]
        )
        out_of_plane_joint_range = jnp.array(
            [joint.range for joint in out_of_plane_joints]
        ).T
        in_plane_joint_qvel_adr = jnp.array(
            [joint.dofadr[0] for joint in in_plane_joints]
        )
        out_of_plane_joint_qvel_adr = jnp.array(
            [joint.dofadr[0] for joint in out_of_plane_joints]
        )

        in_plane_joint_position_observable = MJXObservable(
            name="in_plane_joint_position",
            low=in_plane_joint_range[0],
            high=in_plane_joint_range[1],
            retriever=lambda state: state.mjx_data.qpos[in_plane_joint_qpos_adr],
        )

        out_of_plane_joint_position_observable = MJXObservable(
            name="out_of_plane_joint_position",
            low=out_of_plane_joint_range[0],
            high=out_of_plane_joint_range[1],
            retriever=lambda state: state.mjx_data.qpos[out_of_plane_joint_qpos_adr],
        )

        in_plane_joint_velocity_observable = MJXObservable(
            name="in_plane_joint_velocity",
            low=-jnp.inf * jnp.ones(len(in_plane_joints)),
            high=jnp.inf * jnp.ones(len(in_plane_joints)),
            retriever=lambda state: state.mjx_data.qvel[in_plane_joint_qvel_adr],
        )
        out_of_plane_joint_velocity_observable = MJXObservable(
            name="out_of_plane_joint_velocity",
            low=-jnp.inf * jnp.ones(len(out_of_plane_joints)),
            high=jnp.inf * jnp.ones(len(out_of_plane_joints)),
            retriever=lambda state: state.mjx_data.qvel[out_of_plane_joint_qvel_adr],
        )

        segment_capsule_geom_ids = jnp.array(
            [
                geom_id
                for geom_id in range(self.frozen_mj_model.ngeom)
                if "segment" in self.frozen_mj_model.geom(geom_id).name
                and "capsule" in self.frozen_mj_model.geom(geom_id).name
            ]
        )
        ground_floor_geom_id = self.frozen_mj_model.geom("groundplane").id

        def get_segment_ground_contacts(state: MJXEnvState) -> jnp.ndarray:
            contact_data = state.mjx_data.contact
            contacts = contact_data.dist <= 0
            valid_geom1s = contact_data.geom1 == ground_floor_geom_id

            def solve_contact(geom_id: int) -> jnp.ndarray:
                return (
                    jnp.sum(contacts * valid_geom1s * (contact_data.geom2 == geom_id))
                    > 0
                ).astype(int)

            return jax.vmap(solve_contact)(segment_capsule_geom_ids)

        touch_observable = MJXObservable(
            name="segment_ground_contact",
            low=jnp.zeros(len(segment_capsule_geom_ids)),
            high=jnp.ones(len(segment_capsule_geom_ids)),
            retriever=get_segment_ground_contacts,
        )

        # torso framequat
        torso_id = self.frozen_mj_model.body("ToyExampleMorphology/torso").id
        torso_rotation_observable = MJXObservable(
            name="torso_rotation",
            low=-jnp.pi * jnp.ones(3),
            high=jnp.pi * jnp.ones(3),
            retriever=lambda state: Rotation.from_quat(
                quat=state.mjx_data.xquat[torso_id]
            ).as_euler(seq="xyz"),
        )

        # torso framelinvel
        morphology_freejoint_adr = self.frozen_mj_model.joint(
            "ToyExampleMorphology/freejoint/"
        ).dofadr[0]
        torso_linvel_observable = MJXObservable(
            name="torso_linear_velocity",
            low=-jnp.inf * jnp.ones(3),
            high=jnp.inf * jnp.ones(3),
            retriever=lambda state: state.mjx_data.qvel[
                morphology_freejoint_adr : morphology_freejoint_adr + 3
            ],
        )
        # torso frameangvel
        torso_angvel_observable = MJXObservable(
            name="torso_angular_velocity",
            low=-jnp.inf * jnp.ones(3),
            high=jnp.inf * jnp.ones(3),
            retriever=lambda state: state.mjx_data.qvel[
                morphology_freejoint_adr + 3 : morphology_freejoint_adr + 6
            ],
        )

        # direction to target
        unit_xy_direction_to_target_observable = MJXObservable(
            name="unit_xy_direction_to_target",
            low=-jnp.ones(2),
            high=jnp.ones(2),
            retriever=lambda state: self._get_xy_direction_to_target(state=state)
            / self._get_xy_distance_to_target(state=state),
        )
        # distance to target
        xy_distance_to_target_observable = MJXObservable(
            name="xy_distance_to_target",
            low=jnp.zeros(1),
            high=jnp.inf * jnp.ones(1),
            retriever=lambda state: self._get_xy_distance_to_target(state=state),
        )

        return [
            in_plane_joint_position_observable,
            out_of_plane_joint_position_observable,
            in_plane_joint_velocity_observable,
            out_of_plane_joint_velocity_observable,
            touch_observable,
            torso_rotation_observable,
            torso_linvel_observable,
            torso_angvel_observable,
            unit_xy_direction_to_target_observable,
            xy_distance_to_target_observable,
        ]

    def _update_info(self, state: MJXEnvState) -> MJXEnvState:
        info = {"time": state.mjx_data.time}

        return state.replace(info=info)

    def _update_reward(
        self, state: MJXEnvState, previous_state: MJXEnvState
    ) -> MJXEnvState:
        current_distance_to_target = self._get_xy_distance_to_target(state=state)
        previous_distance_to_target = self._get_xy_distance_to_target(
            state=previous_state
        )
        reward = previous_distance_to_target - current_distance_to_target

        return state.replace(reward=reward)

    def _update_terminated(self, state: MJXEnvState) -> bool:
        terminated = self._get_xy_distance_to_target(state=state) < 0.2

        return state.replace(terminated=terminated)

    def _update_truncated(self, state: MJXEnvState) -> bool:
        truncated = state.mjx_data.time > self.environment_configuration.simulation_time

        return state.replace(truncated=truncated)

    def _get_random_target_position(self, rng: jnp.ndarray) -> jnp.ndarray:
        angle = jax.random.uniform(key=rng, shape=(), minval=0, maxval=jnp.pi * 2)
        radius = self.environment_configuration.target_distance
        target_pos = jnp.array([radius * jnp.cos(angle), radius * jnp.sin(angle), 0.05])
        return target_pos

    def reset(self, rng: chex.PRNGKey) -> MJXEnvState:
        rng, target_pos_rng, qpos_rng, qvel_rng = jax.random.split(key=rng, num=4)

        (mj_model, mj_data), (mjx_model, mjx_data) = self._prepare_reset()

        target_body_id = mj_model.body("target").id
        torso_body_id = mj_model.body("ToyExampleMorphology/torso").id

        # Set random target position
        target_pos = self._get_random_target_position(rng=target_pos_rng)
        mjx_model = mjx_model.replace(
            body_pos=mjx_model.body_pos.at[target_body_id].set(target_pos)
        )

        # Set morphology position
        morphology_pos = jnp.array([0.0, 0.0, 0.11])
        mjx_model = mjx_model.replace(
            body_pos=mjx_model.body_pos.at[torso_body_id].set(morphology_pos)
        )

        # Add noise to initial qpos and qvel of segment joints
        qpos = jnp.copy(mjx_model.qpos0)
        qvel = jnp.zeros(mjx_model.nv)

        qpos = qpos.at[self._segment_joint_qpos_adrs].set(
            qpos[self._segment_joint_qpos_adrs]
            + jax.random.uniform(
                key=qpos_rng,
                shape=(len(self._segment_joints),),
                minval=-self.environment_configuration.joint_randomization_noise_scale,
                maxval=self.environment_configuration.joint_randomization_noise_scale,
            )
        )
        qvel = qvel.at[self._segment_joint_qvel_adrs].set(
            jax.random.uniform(
                key=qvel_rng,
                shape=(len(self._segment_joints),),
                minval=-self.environment_configuration.joint_randomization_noise_scale,
                maxval=self.environment_configuration.joint_randomization_noise_scale,
            )
        )
        mjx_data = mjx_data.replace(qpos=qpos, qvel=qvel, ctrl=jnp.zeros(mjx_model.nu))

        state = self._finish_reset(
            models_and_datas=((mj_model, mj_data), (mjx_model, mjx_data)), rng=rng
        )
        return state



from typing import Callable

import jax
import jax.numpy as jnp
from moojoco.environment import mjx_spaces

from biorobot.toy_example.environment.mjc_env import ToyExampleEnvironmentConfiguration
from biorobot.toy_example.environment.mjx_env import ToyExampleMJXEnvironment
from biorobot.toy_example.mjcf.arena.arena import (
    MJCFPlaneWithTargetArena,
    PlaneWithTargetArenaConfiguration,
)
from biorobot.toy_example.mjcf.morphology.morphology import MJCFToyExampleMorphology
from biorobot.toy_example.mjcf.morphology.specification.default import (
    default_toy_example_morphology_specification,
)
from biorobot.toy_example.usage_examples.mjc.example_usage_single import post_render


def create_mjx_environment(
    environment_configuration: ToyExampleEnvironmentConfiguration,
) -> ToyExampleMJXEnvironment:
    morphology_specification = default_toy_example_morphology_specification(
        num_arms=4, num_segments_per_arm=2
    )
    morphology = MJCFToyExampleMorphology(specification=morphology_specification)
    arena_configuration = PlaneWithTargetArenaConfiguration()
    arena = MJCFPlaneWithTargetArena(configuration=arena_configuration)
    env = ToyExampleMJXEnvironment.from_morphology_and_arena(
        morphology=morphology, arena=arena, configuration=environment_configuration
    )
    return env


def create_mjx_open_loop_controller(
    single_action_space: mjx_spaces.Box, num_envs: int
) -> Callable[[float], jnp.ndarray]:
    def open_loop_controller(t: float) -> jnp.ndarray:
        actions = jnp.ones(single_action_space.shape)
        actions = actions.at[jnp.arange(0, len(actions), 2)].set(jnp.cos(5 * t))
        actions = actions.at[jnp.arange(1, len(actions), 2)].set(jnp.sin(5 * t))
        actions = actions.at[jnp.arange(len(actions) // 2, len(actions), 2)].set(
            actions[jnp.arange(len(actions) // 2, len(actions), 2)] * -1
        )
        return actions

    if num_envs > 1:
        open_loop_controller = jax.vmap(open_loop_controller)

    open_loop_controller = jax.jit(open_loop_controller)

    return open_loop_controller


if __name__ == "__main__":
    environment_configuration = ToyExampleEnvironmentConfiguration(
        render_mode="human", camera_ids=[0, 1]
    )
    env = create_mjx_environment(environment_configuration=environment_configuration)
    jit_reset = jax.jit(env.reset)
    jit_step = jax.jit(env.step)

    rng = jax.random.PRNGKey(0)
    state = jit_reset(rng=rng)

    controller = create_mjx_open_loop_controller(
        single_action_space=env.action_space, num_envs=1
    )

    steps = 0
    fps = 30
    while not (state.truncated | state.terminated):
        t = state.info["time"]
        actions = controller(t)

        state = jit_step(state=state, action=actions)
        if steps % int((1 / fps) / environment_configuration.control_timestep) == 0:
            post_render(
                env.render(state), environment_configuration=environment_configuration
            )
            print(state.observations["segment_ground_contact"])

        steps += 1
    env.close()



from itertools import count
from typing import Any, Dict, List

import mujoco
import numpy as np
from moojoco.environment.base import MuJoCoEnvironmentConfiguration
from moojoco.environment.mjc_env import MJCEnv, MJCEnvState, MJCObservable
from transforms3d.euler import quat2euler

from biorobot.toy_example.mjcf.arena.arena import MJCFPlaneWithTargetArena
from biorobot.toy_example.mjcf.morphology.morphology import MJCFToyExampleMorphology


class ToyExampleEnvironmentConfiguration(MuJoCoEnvironmentConfiguration):
    def __init__(
        self,
        target_distance: float = 3,
        joint_randomization_noise_scale: float = 0.0,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.target_distance = target_distance
        self.joint_randomization_noise_scale = joint_randomization_noise_scale


class ToyExampleMJCEnvironment(MJCEnv):
    metadata = {"render_modes": ["human", "rgb_array"]}

    def __init__(
        self,
        configuration: ToyExampleEnvironmentConfiguration,
        mjcf_str: str,
        mjcf_assets: Dict[str, Any],
    ) -> None:
        super().__init__(
            mjcf_str=mjcf_str, mjcf_assets=mjcf_assets, configuration=configuration
        )

    @classmethod
    def from_morphology_and_arena(
        cls,
        morphology: MJCFToyExampleMorphology,
        arena: MJCFPlaneWithTargetArena,
        configuration: ToyExampleEnvironmentConfiguration,
    ) -> MJCEnv:
        return super().from_morphology_and_arena(
            morphology=morphology, arena=arena, configuration=configuration
        )

    @property
    def environment_configuration(self) -> ToyExampleEnvironmentConfiguration:
        return super().environment_configuration

    @staticmethod
    def _get_xy_direction_to_target(state: MJCEnvState) -> np.ndarray:
        target_position = state.mj_data.body("target").xpos
        torso_position = state.mj_data.body("ToyExampleMorphology/torso").xpos
        direction_to_target = target_position - torso_position
        return direction_to_target[:2]

    @staticmethod
    def _get_xy_distance_to_target(state: MJCEnvState) -> float:
        xy_direction_to_target = ToyExampleMJCEnvironment._get_xy_direction_to_target(
            state=state
        )
        xy_distance_to_target = np.linalg.norm(xy_direction_to_target)
        return xy_distance_to_target

    def _create_observables(self) -> List[MJCObservable]:
        sensors = [
            self.frozen_mj_model.sensor(i) for i in range(self.frozen_mj_model.nsensor)
        ]

        # All joint positions
        joint_pos_sensors = [
            sensor
            for sensor in sensors
            if sensor.type[0] == mujoco.mjtSensor.mjSENS_JOINTPOS
        ]
        in_plane_joint_pos_sensors = [
            sensor for sensor in joint_pos_sensors if "in_plane_joint" in sensor.name
        ]
        in_plane_joints = [
            self.frozen_mj_model.joint(sensor.objid[0])
            for sensor in in_plane_joint_pos_sensors
        ]
        out_of_plane_joint_pos_sensors = [
            sensor
            for sensor in joint_pos_sensors
            if "out_of_plane_joint" in sensor.name
        ]
        out_of_plane_joints = [
            self.frozen_mj_model.joint(sensor.objid[0])
            for sensor in out_of_plane_joint_pos_sensors
        ]

        in_plane_joint_pos_observable = MJCObservable(
            name="in_plane_joint_position",
            low=np.array([joint.range[0] for joint in in_plane_joints]),
            high=np.array([joint.range[1] for joint in in_plane_joints]),
            retriever=lambda state: np.array(
                [
                    state.mj_data.sensordata[
                        sensor.adr[0] : sensor.adr[0] + sensor.dim[0]
                    ]
                    for sensor in in_plane_joint_pos_sensors
                ]
            ).flatten(),
        )
        out_of_plane_joint_pos_observable = MJCObservable(
            name="out_of_plane_joint_position",
            low=np.array([joint.range[0] for joint in out_of_plane_joints]),
            high=np.array([joint.range[1] for joint in out_of_plane_joints]),
            retriever=lambda state: np.array(
                [
                    state.mj_data.sensordata[
                        sensor.adr[0] : sensor.adr[0] + sensor.dim[0]
                    ]
                    for sensor in out_of_plane_joint_pos_sensors
                ]
            ).flatten(),
        )

        # All joint velocities
        joint_vel_sensors = [
            sensor
            for sensor in sensors
            if sensor.type[0] == mujoco.mjtSensor.mjSENS_JOINTVEL
        ]
        in_plane_joint_velocity_sensors = [
            sensor for sensor in joint_vel_sensors if "in_plane_joint" in sensor.name
        ]
        out_of_plane_joint_velocity_sensors = [
            sensor
            for sensor in joint_vel_sensors
            if "out_of_plane_joint" in sensor.name
        ]
        in_plane_joint_velocity_observable = MJCObservable(
            name="in_plane_joint_velocity",
            low=-np.inf * np.ones(len(in_plane_joint_velocity_sensors)),
            high=np.inf * np.ones(len(in_plane_joint_velocity_sensors)),
            retriever=lambda state: np.array(
                [
                    state.mj_data.sensordata[
                        sensor.adr[0] : sensor.adr[0] + sensor.dim[0]
                    ]
                    for sensor in in_plane_joint_velocity_sensors
                ]
            ).flatten(),
        )
        out_of_plane_joint_velocity_observable = MJCObservable(
            name="out_of_plane_joint_velocity",
            low=-np.inf * np.ones(len(out_of_plane_joint_velocity_sensors)),
            high=np.inf * np.ones(len(out_of_plane_joint_velocity_sensors)),
            retriever=lambda state: np.array(
                [
                    state.mj_data.sensordata[
                        sensor.adr[0] : sensor.adr[0] + sensor.dim[0]
                    ]
                    for sensor in out_of_plane_joint_velocity_sensors
                ]
            ).flatten(),
        )

        # segment touch values
        indexer = count(0)
        segment_capsule_geom_ids_to_contact_idx = {
            geom_id: next(indexer)
            for geom_id in range(self.frozen_mj_model.ngeom)
            if "segment" in self.frozen_mj_model.geom(geom_id).name
            and "capsule" in self.frozen_mj_model.geom(geom_id).name
        }
        ground_floor_geom_id = self.frozen_mj_model.geom("groundplane").id

        def get_segment_ground_contacts(state: MJCEnvState) -> np.ndarray:
            ground_contacts = np.zeros(len(segment_capsule_geom_ids_to_contact_idx))
            # based on https://gist.github.com/WuXinyang2012/b6649817101dfcb061eff901e9942057
            for contact_id in range(state.mj_data.ncon):
                contact = state.mj_data.contact[contact_id]
                if contact.geom1 == ground_floor_geom_id:
                    if contact.geom2 in segment_capsule_geom_ids_to_contact_idx:
                        c_array = np.zeros(6, dtype=np.float64)
                        mujoco.mj_contactForce(
                            m=state.mj_model,
                            d=state.mj_data,
                            id=contact_id,
                            result=c_array,
                        )

                        # Convert the contact force from contact frame to world frame
                        ref = np.reshape(contact.frame, (3, 3))
                        c_force = np.dot(np.linalg.inv(ref), c_array[0:3])

                        index = segment_capsule_geom_ids_to_contact_idx[contact.geom2]
                        ground_contacts[index] = max(
                            np.linalg.norm(c_force), ground_contacts[index]
                        )

            ground_contacts = (ground_contacts > 0).astype(int)
            return ground_contacts

        touch_observable = MJCObservable(
            name="segment_ground_contact",
            low=np.zeros(len(segment_capsule_geom_ids_to_contact_idx)),
            high=np.ones(len(segment_capsule_geom_ids_to_contact_idx)),
            retriever=get_segment_ground_contacts,
        )

        # torso framequat
        framequat_sensor = [
            sensor
            for sensor in sensors
            if sensor.type[0] == mujoco.mjtSensor.mjSENS_FRAMEQUAT
        ][0]
        torso_rotation_observable = MJCObservable(
            name="torso_rotation",
            low=-np.pi * np.ones(3),
            high=np.pi * np.ones(3),
            retriever=lambda state: np.array(
                quat2euler(
                    state.mj_data.sensordata[
                        framequat_sensor.adr[0] : framequat_sensor.adr[0]
                        + framequat_sensor.dim[0]
                    ]
                )
            ),
        )

        # torso framelinvel
        framelinvel_sensor = [
            sensor
            for sensor in sensors
            if sensor.type[0] == mujoco.mjtSensor.mjSENS_FRAMELINVEL
        ][0]
        torso_linvel_observable = MJCObservable(
            name="torso_linear_velocity",
            low=-np.inf * np.ones(3),
            high=np.inf * np.ones(3),
            retriever=lambda state: np.array(
                state.mj_data.sensordata[
                    framelinvel_sensor.adr[0] : framelinvel_sensor.adr[0]
                    + framelinvel_sensor.dim[0]
                ]
            ),
        )

        # torso frameangvel
        frameangvel_sensor = [
            sensor
            for sensor in sensors
            if sensor.type[0] == mujoco.mjtSensor.mjSENS_FRAMEANGVEL
        ][0]
        torso_angvel_observable = MJCObservable(
            name="torso_angular_velocity",
            low=-np.inf * np.ones(3),
            high=np.inf * np.ones(3),
            retriever=lambda state: np.array(
                state.mj_data.sensordata[
                    frameangvel_sensor.adr[0] : frameangvel_sensor.adr[0]
                    + frameangvel_sensor.dim[0]
                ]
            ),
        )

        # direction to target
        unit_xy_direction_to_target_observable = MJCObservable(
            name="unit_xy_direction_to_target",
            low=-np.ones(2),
            high=np.ones(2),
            retriever=lambda state: self._get_xy_direction_to_target(state=state)
            / self._get_xy_distance_to_target(state=state),
        )

        # distance to target
        xy_distance_to_target_observable = MJCObservable(
            name="xy_distance_to_target",
            low=np.zeros(1),
            high=np.inf * np.ones(1),
            retriever=lambda state: np.array(
                [self._get_xy_distance_to_target(state=state)]
            ),
        )

        return [
            in_plane_joint_pos_observable,
            out_of_plane_joint_pos_observable,
            in_plane_joint_velocity_observable,
            out_of_plane_joint_velocity_observable,
            touch_observable,
            torso_rotation_observable,
            torso_linvel_observable,
            torso_angvel_observable,
            unit_xy_direction_to_target_observable,
            xy_distance_to_target_observable,
        ]

    def _update_info(self, state: MJCEnvState) -> MJCEnvState:
        info = {"time": state.mj_data.time}

        return state.replace(info=info)

    def _update_reward(
        self, state: MJCEnvState, previous_state: MJCEnvState
    ) -> MJCEnvState:
        current_distance_to_target = self._get_xy_distance_to_target(state=state)
        previous_distance_to_target = self._get_xy_distance_to_target(
            state=previous_state
        )
        reward = previous_distance_to_target - current_distance_to_target

        return state.replace(reward=reward)

    def _update_terminated(self, state: MJCEnvState) -> bool:
        terminated = self._get_xy_distance_to_target(state=state) < 0.2

        return state.replace(terminated=terminated)

    def _update_truncated(self, state: MJCEnvState) -> bool:
        truncated = state.mj_data.time > self.environment_configuration.simulation_time

        return state.replace(truncated=truncated)

    def reset(self, rng: np.random.RandomState) -> MJCEnvState:
        mj_model, mj_data = self._prepare_reset()

        # Set random target position
        angle = rng.uniform(0, 2 * np.pi)
        radius = self.environment_configuration.target_distance

        mj_model.body("target").pos = [
            radius * np.cos(angle),
            radius * np.sin(angle),
            0.05,
        ]

        # Set morphology position
        mj_model.body("ToyExampleMorphology/torso").pos[2] = 0.11

        # Add noise to initial qpos and qvel of segment joints
        segment_joints = [
            mj_model.joint(joint_id)
            for joint_id in range(mj_model.njnt)
            if "segment" in mj_model.joint(joint_id).name
        ]
        num_segment_joints = len(segment_joints)
        joint_qpos_adrs = [joint.qposadr[0] for joint in segment_joints]
        mj_data.qpos[joint_qpos_adrs] = mj_model.qpos0[joint_qpos_adrs] + rng.uniform(
            low=-self.environment_configuration.joint_randomization_noise_scale,
            high=self.environment_configuration.joint_randomization_noise_scale,
            size=num_segment_joints,
        )
        joint_qvel_adrs = [joint.dofadr[0] for joint in segment_joints]
        mj_data.qvel[joint_qvel_adrs] = rng.uniform(
            low=-self.environment_configuration.joint_randomization_noise_scale,
            high=self.environment_configuration.joint_randomization_noise_scale,
            size=num_segment_joints,
        )

        state = self._finish_reset(models_and_datas=(mj_model, mj_data), rng=rng)
        return state



from typing import List

import jax
import jax.numpy as jnp
import mujoco
from jax._src.scipy.spatial.transform import Rotation
from moojoco.environment.mjx_env import MJXEnvState, MJXObservable

"""
This will be merged with the MJC observables once MJX supports sensors.
"""


def get_shared_brittle_star_mjx_observables(
    mj_model: mujoco.MjModel, mj_data: mujoco.MjData
) -> List[MJXObservable]:
    joints = [
        mj_model.joint(joint_id)
        for joint_id in range(mj_model.njnt)
        if "segment" in mj_model.joint(joint_id).name
    ]
    joints_qpos_adr = jnp.array([joint.qposadr[0] for joint in joints])
    joint_dof_adr = jnp.array([joint.dofadr[0] for joint in joints])
    joint_range = jnp.array([joint.range for joint in joints]).T

    joint_position_observable = MJXObservable(
        name="joint_position",
        low=joint_range[0],
        high=joint_range[1],
        retriever=lambda state: state.mjx_data.qpos[joints_qpos_adr],
    )

    joint_velocity_observable = MJXObservable(
        name="joint_velocity",
        low=-jnp.inf * jnp.ones(len(joints)),
        high=jnp.inf * jnp.ones(len(joints)),
        retriever=lambda state: state.mjx_data.qvel[joint_dof_adr],
    )

    actuator_force_observable = MJXObservable(
        name="joint_actuator_force",
        low=-jnp.inf * jnp.ones(len(joint_dof_adr)),
        high=jnp.inf * jnp.ones(len(joint_dof_adr)),
        retriever=lambda state: state.mjx_data.qfrc_actuator[joint_dof_adr].flatten(),
    )

    segment_capsule_geom_ids = jnp.array(
        [
            geom_id
            for geom_id in range(mj_model.ngeom)
            if "segment" in mj_model.geom(geom_id).name
            and "capsule" in mj_model.geom(geom_id).name
        ]
    )

    external_contact_geom_ids = jnp.array(
        [
            mj_model.geom("groundplane").id,
            mj_model.geom("north_wall").id,
            mj_model.geom("east_wall").id,
            mj_model.geom("south_wall").id,
            mj_model.geom("west_wall").id,
        ]
    )

    def get_segment_contacts(state: MJXEnvState) -> jnp.ndarray:
        contact_data = state.mjx_data.contact
        contacts = contact_data.dist <= 0
        valid_geom1s = jnp.isin(contact_data.geom1, external_contact_geom_ids)

        def solve_contact(geom_id: int) -> jnp.ndarray:
            return (
                jnp.sum(contacts * valid_geom1s * (contact_data.geom2 == geom_id)) > 0
            ).astype(int)

        return jax.vmap(solve_contact)(segment_capsule_geom_ids)

    touch_observable = MJXObservable(
        name="segment_contact",
        low=jnp.zeros(len(segment_capsule_geom_ids)),
        high=jnp.ones(len(segment_capsule_geom_ids)),
        retriever=get_segment_contacts,
    )

    disk_id = mj_model.body("BrittleStarMorphology/central_disk").id
    disk_position_observable = MJXObservable(
        name="disk_position",
        low=-jnp.inf * jnp.ones(3),
        high=jnp.inf * jnp.ones(3),
        retriever=lambda state: state.mjx_data.xpos[disk_id],
    )

    # disk framequat
    disk_rotation_observable = MJXObservable(
        name="disk_rotation",
        low=-jnp.pi * jnp.ones(3),
        high=jnp.pi * jnp.ones(3),
        retriever=lambda state: Rotation.from_quat(
            quat=state.mjx_data.xquat[disk_id]
        ).as_euler(seq="xyz"),
    )

    # disk framelinvel
    morphology_freejoint_adr = mj_model.joint(
        "BrittleStarMorphology/freejoint/"
    ).dofadr[0]
    disk_linvel_observable = MJXObservable(
        name="disk_linear_velocity",
        low=-jnp.inf * jnp.ones(3),
        high=jnp.inf * jnp.ones(3),
        retriever=lambda state: state.mjx_data.qvel[
            morphology_freejoint_adr : morphology_freejoint_adr + 3
        ],
    )
    # disk frameangvel
    disk_angvel_observable = MJXObservable(
        name="disk_angular_velocity",
        low=-jnp.inf * jnp.ones(3),
        high=jnp.inf * jnp.ones(3),
        retriever=lambda state: state.mjx_data.qvel[
            morphology_freejoint_adr + 3 : morphology_freejoint_adr + 6
        ],
    )

    return [
        joint_position_observable,
        joint_velocity_observable,
        actuator_force_observable,
        touch_observable,
        disk_position_observable,
        disk_rotation_observable,
        disk_linvel_observable,
        disk_angvel_observable,
    ]



import numpy as np
from moojoco.mjcf.morphology import MJCFMorphology
from transforms3d.euler import euler2quat

from biorobot.toy_example.mjcf.morphology.parts.arm import MJCFToyExampleArm
from biorobot.toy_example.mjcf.morphology.parts.torso import MJCFToyExampleTorso
from biorobot.toy_example.mjcf.morphology.specification.specification import (
    ToyExampleMorphologySpecification,
)


class MJCFToyExampleMorphology(MJCFMorphology):
    def __init__(self, specification: ToyExampleMorphologySpecification) -> None:
        super().__init__(specification=specification, name="ToyExampleMorphology")

    @property
    def morphology_specification(self) -> ToyExampleMorphologySpecification:
        return super().morphology_specification

    def _build(self, *args, **kwargs) -> None:
        self._configure_compiler()
        self._configure_defaults()

        self._build_torso()
        self._build_arms()

        self._configure_camera()

    def _configure_compiler(self) -> None:
        self.mjcf_model.compiler.angle = "radian"

    def _configure_defaults(self) -> None:
        self.mjcf_model.default.geom.condim = 3
        self.mjcf_model.default.geom.friction = [1.0, 0.5, 0.5]
        self.mjcf_model.default.geom.contype = 1
        self.mjcf_model.default.geom.conaffinity = 0

    def _build_torso(self) -> None:
        self.torso = MJCFToyExampleTorso(
            parent=self, name="torso", pos=np.zeros(3), euler=np.zeros(3)
        )

    def _build_arms(self) -> None:
        # Equally spaced over the torso
        self.arms = []

        torso_radius = self.morphology_specification.torso_specification.radius.value
        number_of_arms = len(self.morphology_specification.arm_specifications)
        arm_angles = np.linspace(0, 2 * np.pi, number_of_arms + 1)[:-1]

        for arm_index in range(number_of_arms):
            angle = arm_angles[arm_index]
            position = torso_radius * np.array([np.cos(angle), np.sin(angle), 0])
            arm = MJCFToyExampleArm(
                parent=self.torso,
                name=f"arm_{arm_index}",
                pos=position,
                euler=[0, 0, angle],
                arm_index=arm_index,
            )
            self.arms.append(arm)

    def _configure_camera(self) -> None:
        self.torso.mjcf_body.add(
            "camera",
            name="side_camera",
            pos=[0.0, -2.0, 2.5],
            quat=euler2quat(40 / 180 * np.pi, 0, 0),
            mode="track",
        )



from typing import Tuple

from moojoco.mjcf.arena import ArenaConfiguration, MJCFArena
from transforms3d.euler import euler2quat

from biorobot.toy_example.mjcf.arena.target import MJCFTarget


class PlaneWithTargetArenaConfiguration(ArenaConfiguration):
    def __init__(self, size: Tuple[int, int] = (8, 8)) -> None:
        super().__init__(name="PlaneWithTargetArena")
        self.size = size


class MJCFPlaneWithTargetArena(MJCFArena):
    @property
    def arena_configuration(self) -> PlaneWithTargetArenaConfiguration:
        return super().arena_configuration

    def _build(
        self,
    ) -> None:
        self._build_floor()
        self._configure_lights()
        self._configure_top_camera()
        self.target = self._attach_target()

    def _build_floor(self) -> None:
        self._ground_texture = self.mjcf_model.asset.add(
            "texture",
            rgb1=[0.2, 0.3, 0.4],
            rgb2=[0.1, 0.2, 0.3],
            type="2d",
            builtin="checker",
            name="groundplane",
            width=200,
            height=200,
            mark="edge",
            markrgb=[0.8, 0.8, 0.8],
        )
        self._ground_material = self.mjcf_model.asset.add(
            "material",
            name="groundplane",
            texrepeat=[2, 2],  # Makes white squares exactly 1x1 length units.
            texuniform=True,
            reflectance=0.2,
            texture=self._ground_texture,
        )
        # Build groundplane.
        self._ground_geom = self.mjcf_body.add(
            "geom",
            type="plane",
            name="groundplane",
            material=self._ground_material,
            size=list(self.arena_configuration.size) + [0.25],
            condim=3,
            friction=[1, 0.5, 0.5],
            contype=0,
            conaffinity=1,
        )

    def _configure_lights(self) -> None:
        self.mjcf_model.visual.headlight.set_attributes(
            ambient=[0.4, 0.4, 0.4], diffuse=[0.8, 0.8, 0.8], specular=[0.1, 0.1, 0.1]
        )

    def _configure_top_camera(self) -> None:
        self._main_camera = self.mjcf_body.add(
            "camera",
            name="main_camera",
            pos=[0, 0, 10],
            quat=euler2quat(0, 0, 0),
        )

    def _attach_target(self) -> MJCFTarget:
        target = MJCFTarget(parent=self, name="target")
        return target



from typing import Callable, List

import cv2
import gymnasium
import numpy as np

from biorobot.toy_example.environment.mjc_env import (
    ToyExampleEnvironmentConfiguration,
    ToyExampleMJCEnvironment,
)
from biorobot.toy_example.mjcf.arena.arena import (
    MJCFPlaneWithTargetArena,
    PlaneWithTargetArenaConfiguration,
)
from biorobot.toy_example.mjcf.morphology.morphology import MJCFToyExampleMorphology
from biorobot.toy_example.mjcf.morphology.specification.default import (
    default_toy_example_morphology_specification,
)


def create_mjc_environment(
    environment_configuration: ToyExampleEnvironmentConfiguration,
) -> ToyExampleMJCEnvironment:
    morphology_specification = default_toy_example_morphology_specification(
        num_arms=4, num_segments_per_arm=2
    )
    morphology = MJCFToyExampleMorphology(specification=morphology_specification)
    arena_configuration = PlaneWithTargetArenaConfiguration()
    arena = MJCFPlaneWithTargetArena(configuration=arena_configuration)

    env = ToyExampleMJCEnvironment.from_morphology_and_arena(
        morphology=morphology, arena=arena, configuration=environment_configuration
    )
    return env


def create_mjc_open_loop_controller(
    single_action_space: gymnasium.spaces.Box, num_envs: int
) -> Callable[[float], np.ndarray]:
    def open_loop_controller(t: float) -> np.ndarray:
        actions = np.ones(single_action_space.shape)
        actions[::2] = np.cos(5 * t)
        actions[1::2] = np.sin(5 * t)
        actions[-actions.shape[0] // 2 :: 2] *= -1
        return actions

    if num_envs > 1:
        batched_open_loop_controller = lambda t: np.stack(
            [open_loop_controller(tt) for tt in t]
        )
        return batched_open_loop_controller
    return open_loop_controller


def post_render(
    render_output: List[np.ndarray],
    environment_configuration: ToyExampleEnvironmentConfiguration,
) -> None:
    if environment_configuration.render_mode == "human":
        return

    num_cameras = len(environment_configuration.camera_ids)
    num_envs = len(render_output) // num_cameras

    if num_cameras > 1:
        # Horizontally stack frames of the same environment
        frames_per_env = np.array_split(render_output, num_envs)
        render_output = [
            np.concatenate(env_frames, axis=1) for env_frames in frames_per_env
        ]

    # Vertically stack frames of different environments
    stacked_frames = np.concatenate(render_output, axis=0)
    cv2.imshow("render", stacked_frames)
    cv2.waitKey(1)


if __name__ == "__main__":
    environment_configuration = ToyExampleEnvironmentConfiguration(
        render_mode="human", camera_ids=[0, 1]
    )
    env = create_mjc_environment(environment_configuration=environment_configuration)
    state = env.reset(rng=np.random.RandomState(seed=0))

    controller = create_mjc_open_loop_controller(
        single_action_space=env.action_space, num_envs=1
    )

    steps = 0
    fps = 30
    while not (state.truncated | state.terminated):
        action = controller(state.info["time"])
        state = env.step(state=state, action=action)

        if steps % int((1 / fps) / environment_configuration.control_timestep) == 0:
            post_render(
                env.render(state=state),
                environment_configuration=environment_configuration,
            )
            print(state.observations["segment_ground_contact"])

        steps += 1
    env.close()



import functools

import numpy as np
from moojoco.environment.mjc_env import ThreadedVectorMJCEnvWrapper

from biorobot.toy_example.environment.mjc_env import ToyExampleEnvironmentConfiguration
from biorobot.toy_example.usage_examples.mjc.example_usage_single import (
    create_mjc_environment,
    create_mjc_open_loop_controller,
    post_render,
)


def create_batch_mjc_environment(
    num_envs: int, environment_configuration: ToyExampleEnvironmentConfiguration
) -> ThreadedVectorMJCEnvWrapper:

    batched_env = ThreadedVectorMJCEnvWrapper(
        create_env_fn=functools.partial(
            create_mjc_environment, environment_configuration=environment_configuration
        ),
        num_environments=num_envs,
    )
    return batched_env


if __name__ == "__main__":
    num_envs = 3
    environment_configuration = ToyExampleEnvironmentConfiguration(
        render_mode="rgb_array", camera_ids=[0, 1], joint_randomization_noise_scale=0.01
    )
    batched_env = create_batch_mjc_environment(
        num_envs=num_envs, environment_configuration=environment_configuration
    )
    rng = [np.random.RandomState(i) for i in range(num_envs)]
    batched_state = batched_env.reset(rng=rng)

    controller = create_mjc_open_loop_controller(
        single_action_space=batched_env.single_action_space, num_envs=num_envs
    )

    steps = 0
    fps = 30
    while not np.any(batched_state.terminated | batched_state.truncated):
        ts = batched_state.info["time"]
        batched_action = controller(ts)
        batched_state = batched_env.step(state=batched_state, action=batched_action)

        if steps % int((1 / fps) / environment_configuration.control_timestep) == 0:
            post_render(
                batched_env.render(state=batched_state),
                environment_configuration=environment_configuration,
            )

        steps += 1
    batched_env.close()



import jax
import jax.numpy as jnp

from biorobot.toy_example.environment.mjc_env import ToyExampleEnvironmentConfiguration
from biorobot.toy_example.usage_examples.mjc.example_usage_single import post_render
from biorobot.toy_example.usage_examples.mjx.example_usage_single import (
    create_mjx_environment,
    create_mjx_open_loop_controller,
)

if __name__ == "__main__":
    environment_configuration = ToyExampleEnvironmentConfiguration(
        render_mode="rgb_array", camera_ids=[0, 1], joint_randomization_noise_scale=0.1
    )
    env = create_mjx_environment(environment_configuration=environment_configuration)

    num_envs = 2
    rng = jax.random.PRNGKey(0)
    rng, *subrngs = jax.random.split(key=rng, num=num_envs + 1)

    jit_batched_reset = jax.jit(jax.vmap(env.reset))
    jit_batched_step = jax.jit(jax.vmap(env.step))
    batched_state = jit_batched_reset(rng=jnp.array(subrngs))

    controller = create_mjx_open_loop_controller(
        single_action_space=env.action_space, num_envs=num_envs
    )

    steps = 0
    fps = 30
    while not jnp.any(batched_state.terminated | batched_state.truncated):
        ts = batched_state.info["time"]
        actions = controller(ts)

        batched_state = jit_batched_step(state=batched_state, action=actions)

        if steps % int((1 / fps) / environment_configuration.control_timestep) == 0:
            post_render(
                env.render(batched_state),
                environment_configuration=environment_configuration,
            )
            print(batched_state.observations["segment_ground_contact"])

        steps += 1
    env.close()



import jax.numpy as jnp
import mujoco


class BrittleStarEnvironmentBase:
    def __init__(self) -> None:
        self._segment_joints_qpos_adrs = None
        self._segment_joints_qvel_adrs = None
        self._segment_capsule_geom_ids = None

    def _cache_references(self, mj_model: mujoco.MjModel) -> None:
        self._get_segment_capsule_geom_ids(mj_model=mj_model)
        self._get_segment_joints_qpos_adrs(mj_model=mj_model)
        self._get_segment_joints_qvel_adrs(mj_model=mj_model)

    def _get_segment_joints_qpos_adrs(self, mj_model: mujoco.MjModel) -> jnp.ndarray:
        if self._segment_joints_qpos_adrs is None:
            self._segment_joints_qpos_adrs = jnp.array(
                [
                    mj_model.joint(joint_id).qposadr[0]
                    for joint_id in range(mj_model.njnt)
                    if "segment" in mj_model.joint(joint_id).name
                ]
            )
        return self._segment_joints_qpos_adrs

    def _get_segment_joints_qvel_adrs(self, mj_model: mujoco.MjModel) -> jnp.ndarray:
        if self._segment_joints_qvel_adrs is None:
            self._segment_joints_qvel_adrs = jnp.array(
                [
                    mj_model.joint(joint_id).dofadr[0]
                    for joint_id in range(mj_model.njnt)
                    if "segment" in mj_model.joint(joint_id).name
                ]
            )
        return self._segment_joints_qvel_adrs

    def _get_segment_capsule_geom_ids(self, mj_model: mujoco.MjModel) -> jnp.ndarray:
        if self._segment_capsule_geom_ids is None:
            self._segment_capsule_geom_ids = jnp.array(
                [
                    geom_id
                    for geom_id in range(mj_model.ngeom)
                    if "segment" in mj_model.geom(geom_id).name
                    and "capsule" in mj_model.geom(geom_id).name
                ]
            )
        return self._segment_capsule_geom_ids



from itertools import count
from typing import List

import mujoco
import numpy as np
from moojoco.environment.mjc_env import MJCEnvState, MJCObservable
from transforms3d.euler import quat2euler


def get_shared_brittle_star_mjc_observables(
    mj_model: mujoco.MjModel, mj_data: mujoco.MjData
) -> List[MJCObservable]:
    sensors = [mj_model.sensor(i) for i in range(mj_model.nsensor)]

    # All joint positions
    joint_pos_sensors = [
        sensor
        for sensor in sensors
        if sensor.type[0] == mujoco.mjtSensor.mjSENS_JOINTPOS
    ]
    joints = [mj_model.joint(sensor.objid[0]) for sensor in joint_pos_sensors]

    joint_position_observable = MJCObservable(
        name="joint_position",
        low=np.array([joint.range[0] for joint in joints]),
        high=np.array([joint.range[1] for joint in joints]),
        retriever=lambda state: np.array(
            [
                state.mj_data.sensordata[sensor.adr[0] : sensor.adr[0] + sensor.dim[0]]
                for sensor in joint_pos_sensors
            ]
        ).flatten(),
    )

    # All joint velocities
    joint_vel_sensors = [
        sensor
        for sensor in sensors
        if sensor.type[0] == mujoco.mjtSensor.mjSENS_JOINTVEL
    ]
    joint_velocity_observable = MJCObservable(
        name="joint_velocity",
        low=-np.inf * np.ones(len(joint_vel_sensors)),
        high=np.inf * np.ones(len(joint_vel_sensors)),
        retriever=lambda state: np.array(
            [
                state.mj_data.sensordata[sensor.adr[0] : sensor.adr[0] + sensor.dim[0]]
                for sensor in joint_vel_sensors
            ]
        ).flatten(),
    )

    # All actuator torques
    actuator_frc_sensors = [
        sensor
        for sensor in sensors
        if sensor.type[0] == mujoco.mjtSensor.mjSENS_JOINTACTFRC
    ]
    actuator_force_observable = MJCObservable(
        name="joint_actuator_force",
        low=-np.inf * np.ones(len(actuator_frc_sensors)),
        high=np.inf * np.ones(len(actuator_frc_sensors)),
        retriever=lambda state: np.array(
            [
                state.mj_data.sensordata[sensor.adr[0] : sensor.adr[0] + sensor.dim[0]]
                for sensor in actuator_frc_sensors
            ]
        ).flatten(),
    )

    # segment touch values
    #   Start by mapping geom indices of segment capsules to a contact output index
    indexer = count(0)
    segment_capsule_geom_id_to_contact_idx = {}
    for geom_id in range(mj_model.ngeom):
        geom_name = mj_model.geom(geom_id).name
        if "segment" in geom_name and "capsule" in geom_name:
            segment_capsule_geom_id_to_contact_idx[geom_id] = next(indexer)

    #   Get non-morphology geom ids to check contacts with
    external_contact_geom_ids = {
        mj_model.geom("groundplane").id,
        mj_model.geom("north_wall").id,
        mj_model.geom("east_wall").id,
        mj_model.geom("south_wall").id,
        mj_model.geom("west_wall").id,
    }

    def get_segment_contacts(state: MJCEnvState) -> np.ndarray:
        contacts = np.zeros(len(segment_capsule_geom_id_to_contact_idx))
        # based on https://gist.github.com/WuXinyang2012/b6649817101dfcb061eff901e9942057
        for contact_id in range(state.mj_data.ncon):
            contact = state.mj_data.contact[contact_id]
            if contact.geom1 in external_contact_geom_ids:
                if contact.geom2 in segment_capsule_geom_id_to_contact_idx:
                    c_array = np.zeros(6, dtype=np.float64)
                    mujoco.mj_contactForce(
                        m=state.mj_model, d=state.mj_data, id=contact_id, result=c_array
                    )

                    # Convert the contact force from contact frame to world frame
                    ref = np.reshape(contact.frame, (3, 3))
                    c_force = np.dot(np.linalg.inv(ref), c_array[0:3])

                    index = segment_capsule_geom_id_to_contact_idx[contact.geom2]
                    contacts[index] = max(np.linalg.norm(c_force), contacts[index])

        ground_contacts = (contacts > 0).astype(int)
        return ground_contacts

    touch_observable = MJCObservable(
        name="segment_contact",
        low=np.zeros(len(segment_capsule_geom_id_to_contact_idx)),
        high=np.ones(len(segment_capsule_geom_id_to_contact_idx)),
        retriever=get_segment_contacts,
    )

    # disk framepos
    framepos_sensor = [
        sensor
        for sensor in sensors
        if sensor.type[0] == mujoco.mjtSensor.mjSENS_FRAMEPOS
    ][0]
    disk_position_observable = MJCObservable(
        name="disk_position",
        low=-np.inf * np.ones(3),
        high=np.inf * np.ones(3),
        retriever=lambda state: np.array(
            state.mj_data.sensordata[
                framepos_sensor.adr[0] : framepos_sensor.adr[0] + framepos_sensor.dim[0]
            ]
        ),
    )

    # disk framequat
    framequat_sensor = [
        sensor
        for sensor in sensors
        if sensor.type[0] == mujoco.mjtSensor.mjSENS_FRAMEQUAT
    ][0]
    disk_rotation_observable = MJCObservable(
        name="disk_rotation",
        low=-np.pi * np.ones(3),
        high=np.pi * np.ones(3),
        retriever=lambda state: np.array(
            quat2euler(
                state.mj_data.sensordata[
                    framequat_sensor.adr[0] : framequat_sensor.adr[0]
                    + framequat_sensor.dim[0]
                ]
            )
        ),
    )

    # disk framelinvel
    framelinvel_sensor = [
        sensor
        for sensor in sensors
        if sensor.type[0] == mujoco.mjtSensor.mjSENS_FRAMELINVEL
    ][0]
    disk_linvel_observable = MJCObservable(
        name="disk_linear_velocity",
        low=-np.inf * np.ones(3),
        high=np.inf * np.ones(3),
        retriever=lambda state: np.array(
            state.mj_data.sensordata[
                framelinvel_sensor.adr[0] : framelinvel_sensor.adr[0]
                + framelinvel_sensor.dim[0]
            ]
        ),
    )

    # disk frameangvel
    frameangvel_sensor = [
        sensor
        for sensor in sensors
        if sensor.type[0] == mujoco.mjtSensor.mjSENS_FRAMEANGVEL
    ][0]
    disk_angvel_observable = MJCObservable(
        name="disk_angular_velocity",
        low=-np.inf * np.ones(3),
        high=np.inf * np.ones(3),
        retriever=lambda state: np.array(
            state.mj_data.sensordata[
                frameangvel_sensor.adr[0] : frameangvel_sensor.adr[0]
                + frameangvel_sensor.dim[0]
            ]
        ),
    )

    return [
        joint_position_observable,
        joint_velocity_observable,
        actuator_force_observable,
        touch_observable,
        disk_position_observable,
        disk_rotation_observable,
        disk_linvel_observable,
        disk_angvel_observable,
    ]
